from pathlib import Path
import torch
from torch import autocast
from diffusers import (
    StableDiffusionImg2ImgPipeline,
    StableDiffusionPipeline,
)

# weird, why didn't this install with vktrs?
#!pip install pytokenizations yt-dlp python-tsp webvtt-py

#use_stability_api = False


# to do: rename "start_schedule" to "strength"
# start_schedule=(1-image_consistency))


class HfHelper:
    def __init__(
        self,
        device = 'cuda',
        device_img2img = None,
        device_text2img = None,
        model_path = '.',
        model_id = "CompVis/stable-diffusion-v1-4",
        download=True,
    ):
        if not device_img2img:
            device_img2img = device
        if not device_text2img:
            device_text2img = device
        self.device = device
        self.device_img2img = device_img2img
        self.device_text2img = device_text2img
        self.model_path = model_path
        self.model_id = model_id
        self.download = download
        self.load_pipelines()

    def load_pipelines(
        self,
    ):
        if self.download:
            img2img = StableDiffusionImg2ImgPipeline.from_pretrained(
                self.model_id,
                revision="fp16", 
                torch_dtype=torch.float16,
                use_auth_token=True
            )
            img2img = img2img.to(self.device)
            img2img.save_pretrained(self.model_path)
        else:
            img2img = StableDiffusionImg2ImgPipeline.from_pretrained(
                self.model_path,
                local_files_only=True
            ).to(self.device)

        text2img = StableDiffusionPipeline(
            vae=img2img.vae,
            text_encoder=img2img.text_encoder,
            tokenizer=img2img.tokenizer,
            unet=img2img.unet,
            feature_extractor=img2img.feature_extractor,
            scheduler=img2img.scheduler,
            safety_checker=img2img.safety_checker,
        )
        #return text2img, img2img
        text2img.enable_attention_slicing()
        img2img.enable_attention_slicing()
        self.text2img = text2img
        self.img2img = img2img

    def get_image_for_prompt(
        self,
        prompt,
        **kwargs
    ):
        f = self.text2img if kwargs.get('init_image') is None else self.img2img
        #if kwargs.get('image_consistency') is not None:
        #kwargs['strength'] = 1- kwargs['image_consistency'] 
        if kwargs.get('start_schedule') is not None:
            #kwargs['strength'] = kwargs['start_schedule']
            kwargs['strength'] = kwargs.pop('start_schedule')
        with autocast(self.device):
            return f(prompt, **kwargs)



import time

import numpy as np
from scipy.spatial.distance import pdist, squareform
from toolz.itertoolz import partition_all
from python_tsp.exact import solve_tsp_dynamic_programming


def tsp_permute_frames(frames, verbose=False):
    """
    Permutes images using traveling salesman solver to find frame-to-frame 
    ordering that minimizes difference between subsequent frames, i.e.
    the ordering of the images that gives the smoothest animation.
    """
    frames_m = np.array([np.array(f).ravel() for f in frames])
    dmat = pdist(frames_m, metric='cosine')
    dmat = squareform(dmat)

    start = time.time()
    permutation, _ = solve_tsp_dynamic_programming(dmat)
    if verbose:
        print(f"elapsed: {time.time() - start}")

    frames_permuted = [frames[i] for i in permutation]
    return frames_permuted

def batched_tsp_permute_frames(frames, batch_size):
    """
    TSP solver is O(n^2). Instead of limiting how many variations a user can
    request for a particular image, we set an upperbound on how many images we
    send to the solver at any given time. TODO: Faster solver.
    """
    ordered = []
    for batch in partition_all(batch_size, frames):
        ordered.extend( tsp_permute_frames(batch) )
    return ordered



from pathlib import Path
import random
import string
import subprocess
import textwrap
import time

import pandas as pd
from PIL import Image, ImageDraw, ImageFont


def gpu_info():
    outv = subprocess.run([
        'nvidia-smi',
            # these lines concatenate into a single query string
            '--query-gpu='
            'timestamp,'
            'name,'
            'utilization.gpu,'
            'utilization.memory,'
            'memory.used,'
            'memory.free,'
            ,
         '--format=csv'
        ],
        stdout=subprocess.PIPE).stdout.decode('utf-8')

    header, rec = outv.split('\n')[:-1]
    return pd.DataFrame({' '.join(k.strip().split('.')).capitalize():v for k,v in zip(header.split(','), rec.split(','))}, index=[0]).T


def get_audio_duration_seconds(audio_fpath):
    outv = subprocess.run([
        'ffprobe'
        ,'-i',audio_fpath
        ,'-show_entries', 'format=duration'
        ,'-v','quiet'
        ,'-of','csv=p=0'
        ],
        stdout=subprocess.PIPE
        ).stdout.decode('utf-8')
    return float(outv.strip())


def rand_str(n_char=5):
    return ''.join(random.choice(string.ascii_lowercase) for i in range(n_char))


def remove_punctuation(s):
    # https://stackoverflow.com/a/266162/819544
    return s.translate(str.maketrans('', '', string.punctuation))


def sanitize_folder_name(fp):
    outv = ''
    whitelist = string.ascii_letters + string.digits + '-_'
    for token in str(fp):
        if token not in whitelist:
            token = '-'
        outv += token
    return outv


def add_caption2image(
      image, 
      caption, 
      text_font='LiberationSans-Regular.ttf', 
      font_size=20,
      fill_color=(255, 255, 255),
      stroke_color=(0, 0, 0), #stroke_fill
      stroke_width=2,
      align='center',
      ):
    # via https://stackoverflow.com/a/59104505/819544
    wrapper = textwrap.TextWrapper(width=50) 
    word_list = wrapper.wrap(text=caption) 
    caption_new = ''
    for ii in word_list[:-1]:
        caption_new = caption_new + ii + '\n'
    caption_new += word_list[-1]

    draw = ImageDraw.Draw(image)

    # Download the Font and Replace the font with the font file. 
    font = ImageFont.truetype(text_font, size=font_size)
    w,h = draw.textsize(caption_new, font=font, stroke_width=stroke_width)
    W,H = image.size
    x,y = 0.5*(W-w),0.90*H-h
    draw.text(
        (x,y), 
        caption_new,
        font=font,
        fill=fill_color, 
        stroke_fill=stroke_color,
        stroke_width=stroke_width,
        align=align,
    )

    return image


def save_frame(
    img: Image,
    idx:int=0,
    root_path=Path('./frames'),
    name=None,
):
    root_path.mkdir(parents=True, exist_ok=True)
    if name is None:
        name = rand_str()
    outpath = root_path / f"{idx}-{name}.png"
    img.save(outpath)
    return str(outpath)


def get_image_sequence(idx, root, init_first=True):
    root = Path(root)
    images = (root / 'frames' ).glob(f'{idx}-*.png')
    images = [str(fp) for fp in images]
    if init_first:
        init_image = None
        images2 = []
        for i, fp in enumerate(images):
            if 'anchor' in fp:
                init_image = fp
            else:
                images2.append(fp)
        if not init_image:
            try:
                init_image, images2 = images2[0], images2[1:]
                images = [init_image] + images2
            except IndexError:
                images = images2
    return images


def archive_images(idx, root, archive_root = None):
    root = Path(root)
    if archive_root is None:
        archive_root = root / 'archive'
    archive_root = Path(archive_root)
    archive_root.mkdir(parents=True, exist_ok=True)
    old_images = get_image_sequence(idx, root=root)
    if not old_images:
        return
    print(f"moving {len(old_images)} old images for scene {idx} to {archive_root}")
    for old_fp in old_images:
        old_fp = Path(old_fp)
        im_name = Path(old_fp.name)
        new_path = archive_root / im_name
        if new_path.exists():
            im_name = f"{im_name.stem}-{time.time()}{im_name.suffix}"
            new_path = archive_root / im_name
        old_fp.rename(new_path)


import io
import os
from PIL import Image, ImageDraw, ImageFont
import warnings

from stability_sdk import client
import stability_sdk.interfaces.gooseai.generation.generation_pb2 as generation


def get_image_for_prompt(prompt, max_retries=3, **kargs):
    stability_api = client.StabilityInference(
        key=os.environ['STABILITY_KEY'], 
        verbose=False,
    )


    # auto-retry if mitigation triggered
    while max_retries:

        try:
            answers = stability_api.generate(prompt=prompt, **kargs)

            response = process_response(answers)

            for img in response:

                yield img

            break # whoops... this breaks us out of the while loop, not the for loop.
        except RuntimeError:
            print("runtime error")
            max_retries -= 1
            warnings.warn(f"mitigation triggered, retries remaining: {max_retries}")


def process_response(answers):

    # iterating over the generator produces the api response
    for resp in answers:

        for artifact in resp.artifacts:

            #print(artifact.finish_reason)
            if artifact.finish_reason == generation.FILTER:

                warnings.warn(
                    "Your request activated the API's safety filters and could not be processed."
                    "Please modify the prompt and try again.")
                #raise RuntimeError
            if artifact.type == generation.ARTIFACT_IMAGE:
                img = Image.open(io.BytesIO(artifact.binary))
                yield img




"""
# Automatic Speech Recognition utilities

Currently uses openai/whisper. To install: 
  pip install git+https://github.com/openai/whisper
"""

import time

import tokenizations
from vktrs.utils import remove_punctuation
import whisper


def whisper_transcribe(
    audio_fpath="audio.mp3",
):
    whispers = {
        'tiny':None, # 5.83 s
        'large':None # 3.73 s
    }
    # accelerated runtime required for whisper
    # to do: pypi package for whisper
    
    for k in whispers.keys():
        options = whisper.DecodingOptions(
            language='en',
        )
        # to do: be more proactive about cleaning up these models when we're done with them
        model = whisper.load_model(k).to('cuda')
        start = time.time()
        print(f"Transcribing audio with whisper-{k}")
        
        # to do: calling transcribe like this unnecessarily re-processes audio each time.
        whispers[k] = model.transcribe(audio_fpath) # re-processes audio each time, ~10s overhead?
        print(f"elapsed: {time.time()-start}")
    return whispers


def whisper_align(whispers):
    # sanitize and tokenize
    whispers_tokens = {}
    for k in whispers:
        whispers_tokens[k] = [
        remove_punctuation(tok) for tok in whispers[k]['text'].split()
        ]

    # align sequences
    tiny2large, large2tiny = tokenizations.get_alignments( #seqdiff.diff( #= tokenizations.get_alignments(
        whispers_tokens['tiny'],
        whispers_tokens['large']
    )
    return tiny2large, large2tiny, whispers_tokens


def whisper_transmit_meta_across_alignment(
    whispers,
    large2tiny,
    whispers_tokens,
):
    idx=0
    tokenized_prompts_tiny = []
    for phrase_idx, phrase in enumerate(whispers['tiny']['segments']):
        rec = {
            'start': phrase['start'],
            'end': phrase['end'],
            'tokens':[],
            'indices':[],
        }

        # TO DO: I should really use this tokenization for the alignment step to ensure the indices match up
        for tok in phrase['text'].split():
            tok = remove_punctuation(tok)
            rec['tokens'].append(tok)
            rec['indices'].append(idx)
            idx+=1

        tokenized_prompts_tiny.append(rec)

    # flatten
    token_tinyindex_segmentations = {}
    for rec in tokenized_prompts_tiny:
        for j, idx in enumerate(rec['indices']):
            token_tinyindex_segmentations[idx] ={
                'token':rec['tokens'][j],
                'start':rec['start'],
                'end':rec['end'],
            }
    
    #token_tinyindex_segmentations
    token_large_index_segmentations = {}
    for i, result in enumerate(large2tiny):
        rec_large = {'token':whispers_tokens['large'][i]}
        for j in result:
            rec_tiny = token_tinyindex_segmentations[j]
            if not rec_large.get('start'):
                rec_large['start'] = rec_tiny['start']
                rec_large['end'] = rec_tiny['end']
        
        # handle null result. this could be more elegant/DRY, but this way is less confusing to me at least.
        # basically, we're just backfilling here, so each entry will have a start and end time
        if not rec_large.get('start'):
            if i == 0:
                rec_large['start'] = 0
            else:
                rec_prev = token_large_index_segmentations[i-1]
                rec_large['start'] = rec_prev['start']
                rec_large['end'] = rec_prev.get('end')
        
        token_large_index_segmentations[i] = rec_large
    
    return token_large_index_segmentations


def whisper_segment_transcription(
    token_large_index_segmentations,
):
    """
    apply whisper-tiny segmentations to whisper-large transcriptions

    """
    token_large_phrase_segmentations = []
    start_prev = 0
    end_prev=0
    current_phrase = []
    for rec in token_large_index_segmentations.values():

        # we're in the same phrase as previous step
        if rec['start'] == start_prev:
            current_phrase.append(rec['token'])
            start_prev = rec['start']
            end_prev = rec.get('end')
            continue

        # we're in the next phrase, 
        token_large_phrase_segmentations.append({
            'tokens': current_phrase,
            'start':start_prev,
            'end':end_prev,
        })
        current_phrase = []

        # ...which starts immediately after the previous phrase
        if rec['start'] == end_prev:
            current_phrase.append(rec['token'])
            start_prev = rec['start']
            end_prev = rec['end']
            continue
        
        # ...or else there's a gap between when the last phrase ended and this one starts,
        # and I'm frankly not sure how I want to adress that yet.
        else:
            #raise NotImplementedError
            # let's just do.. this? for now? I guess?
            current_phrase.append(rec['token'])
            start_prev = rec['start']
            end_prev = rec['end']
            continue

    # how about we not drop the last lyric?
    # should this be a for-finally clause?
    token_large_phrase_segmentations.append({
            'tokens': current_phrase,
            'start':start_prev,
            'end':end_prev,
        })

    # reshape the data structure
    prompt_starts = [
            {'ts':rec['start'],
            'prompt':' '.join(rec['tokens'])
            }
        for rec in token_large_phrase_segmentations]
    
    return prompt_starts


def whisper_lyrics(audio_fpath="audio.mp3"):
    whispers = whisper_transcribe(audio_fpath)
    tiny2large, large2tiny, whispers_tokens = whisper_align(whispers)
    token_large_index_segmentations = whisper_transmit_meta_across_alignment(
        whispers,
        large2tiny,
        whispers_tokens,
    )
    prompt_starts = whisper_segment_transcription(
        token_large_index_segmentations,
    )
    return prompt_starts